/* 
 * Copyright (C) 2006 Cooper Street Innovations Inc.
 *	Charles Eidsness    <charles@cooper-street.com>
 *
 * This program is free software; you can redistribute it and/or
 * modify it under the terms of the GNU General Public License
 * as published by the Free Software Foundation; either version 2
 * of the License, or (at your option) any later version.
 * 
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 * 
 * You should have received a copy of the GNU General Public License
 * along with this program; if not, write to the Free Software
 * Foundation, Inc., 51 Franklin Street, Fifth Floor, Boston, MA 
 * 02110-1301, USA.
 *
 */

#include <math.h>

#include "log.h"
#include "calc.h"
#include "variable.h"
#include "tokenizer.h"
#include "parser.h"

struct _calc {
	variable_ *variables;
	token_ *tokens;
	void *parser;
};

int calcSolve(calc_ *r, double *solution)
{
	token_ *ptr;
	
	ReturnErrIf(r == NULL);
	ReturnErrIf(solution == NULL);
	
	for(ptr = r->tokens; ptr != NULL; ptr = ptr->next) {
		Debug("%s", ParseTokenName(ptr->type));
		Parse(r->parser, ptr->type, ptr, solution);
		ReturnErrIf(isnan(*solution), "Parser failed");
	}
	
	return 0;
}

int calcDiff(calc_ *r, char *variable, char type, double *solution)
{
	token_ *ptr;
	double **diffVariable;
	
	ReturnErrIf(r == NULL);
	ReturnErrIf(solution == NULL);
	
	diffVariable = variableFindData(r->variables, variable, type);
	ReturnErrIf(diffVariable == NULL, "Couldn't find variable %s.", variable);
	
	Debug("%s", ParseTokenName(TOKEN_DIFF));
	Parse(r->parser, TOKEN_DIFF, NULL, solution);
	ReturnErrIf(isnan(*solution), "Parser failed");
	
	for(ptr = r->tokens; ptr != NULL; ptr = ptr->next) {
		if((ptr->type == TOKEN_VARIABLE) && (*ptr->variable == *diffVariable)) {
			Debug("%s", ParseTokenName(TOKEN_DIFF_VARIABLE));
			Parse(r->parser, TOKEN_DIFF_VARIABLE, ptr, solution);
			ReturnErrIf(isnan(*solution), "Parser failed");
		} else {
			Debug("%s", ParseTokenName(ptr->type));
			Parse(r->parser, ptr->type, ptr, solution);
			ReturnErrIf(isnan(*solution), "Parser failed");
		}
	}
	
	return 0;
}

int calcZeroXing(calc_ *r, double *solution)
{
	token_ *ptr;
	
	ReturnErrIf(r == NULL);
	ReturnErrIf(solution == NULL);
	
	for(ptr = r->tokens; ptr != NULL; ptr = ptr->next) {
		if(ptr->type == TOKEN_VARIABLE) {
			Debug("%s", ParseTokenName(TOKEN_ZERO_VARIABLE));
			Parse(r->parser, TOKEN_ZERO_VARIABLE, ptr, solution);
			ReturnErrIf(isnan(*solution), "Parser failed");
		} else {
			Debug("%s", ParseTokenName(ptr->type));
			Parse(r->parser, ptr->type, ptr, solution);
			ReturnErrIf(isnan(*solution), "Parser failed");
		}
	}
	
	return 0;
}

typedef struct {
	void *private;
	calc_ *r;
	calcGetVarPtr getVarFunc;
} getVarFuncArgs;

static double ** localGetVarFunc(char *varName, char type, void *private)
{
	getVarFuncArgs *args;
	double **data;
	variable_ *variable;
	
	args = (getVarFuncArgs *)private;
	
	data = variableFindData(args->r->variables->next, varName, type);
	if(data == NULL) {
		/* variableNew makes a copy of the varible name, we want to pass
			the copy to the user, not the pointer to the equation becasue
			the varible name in the equation is only temperarilly a null
			terminated string */
		variable = variableAttachNew(args->r->variables, varName, type);
		ReturnNULLIf(variable == NULL);
		data = args->getVarFunc(variable->name, type, args->private);
		ReturnNULLIf(data == NULL);
		variable->data = data;
	}
	
	return data;
}

static int init(calc_ *r, char *buffer, calcGetVarPtr getVarFunc, 
		void *private, double *minDiv)
{
	getVarFuncArgs args;
	
	Debug("Creating calculon %p", r);
	
	args.private = private;
	args.r = r;
	args.getVarFunc = getVarFunc;
	
	Debug("%p", &args);
	
	r->variables = variableNew(r->variables);
	ReturnErrIf(r->variables == NULL);
	
	r->tokens = tokenizerNew(r->tokens, buffer, localGetVarFunc, (void *)&args,
		minDiv);
	ReturnErrIf(r->tokens == NULL, "Failed to parse string");
	
	r->parser = ParseAlloc(malloc);
	ReturnErrIf(r->parser == NULL, "Failed to create parser");
	
	return 0;
}

int calcInfo(void)
{
	Info("Calculon %i.%i", CALC_MAJOR_VERSION, CALC_MINOR_VERSION);
	Info("Compiled " __DATE__ " at " __TIME__);
	Info("(c) 2006 Cooper Street Innovations Inc.");
	return ((CALC_MAJOR_VERSION << 16) + CALC_MINOR_VERSION);
}

int calcDestroy(calc_ **r)
{
	ReturnErrIf(r == NULL);
	ReturnErrIf(*r == NULL);
	Debug("Destroying calculon %p", *r);
	
	if((*r)->tokens != NULL)
		tokenizerDestroyAll(&(*r)->tokens);
	if((*r)->parser != NULL)
		ParseFree((*r)->parser, free);
	if((*r)->variables != NULL)
		variableDestroyAll(&(*r)->variables);
	
	free(*r);
	*r = NULL;
	return 0;
}

calc_ * calcNew(calc_ *r, char *buffer, calcGetVarPtr getVarFunc, 
		void *private, double *minDiv)
{
	ReturnNULLIf(NULL != 0);
	ReturnNULLIf(r != NULL);
	ReturnNULLIf(buffer == NULL);
	ReturnNULLIf(getVarFunc == NULL);
	r = calloc(1, sizeof(calc_));
	ReturnNULLIf(r == NULL, "Malloc Failed");
	if(init(r, buffer, getVarFunc, private, minDiv)) {
		calcDestroy(&r);
	}
	return r;
}
