/* 
 * Copyright (C) 2006 Cooper Street Innovations Inc.
 *	Charles Eidsness    <charles@cooper-street.com>
 *
 * This program is free software; you can redistribute it and/or
 * modify it under the terms of the GNU General Public License
 * as published by the Free Software Foundation; either version 2
 * of the License, or (at your option) any later version.
 * 
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 * 
 * You should have received a copy of the GNU General Public License
 * along with this program; if not, write to the Free Software
 * Foundation, Inc., 51 Franklin Street, Fifth Floor, Boston, MA 
 * 02110-1301, USA.
 *
 */

#include "log.h"
#include "tokenizer.h"
#include "parser.h"

/* RE Format parser defines */
#define	YYCTYPE		char
#define	YYCURSOR	cursor
#define	YYMARKER	marker

/*!re2c /*!ignore!re2c Set RE2C Options */
	re2c:yyfill:enable=0;
*/

/*!re2c /*!ignore!re2c Useful Regular Expressions */
	EOI = [\000];
	EOL = [\012] | [\015];
	ANY	= [\000-\377];
	ALPHA	= [a-zA-Z];
	DIGIT	= [0-9];
	NUMBER	= (DIGIT+ | "." DIGIT+ | DIGIT+ "." DIGIT+) ('e'('+'|'-')? DIGIT+)?;
	TEXT = (ALPHA | DIGIT);
	WSP = (" ") | [\011];
	CONTINUE = EOL+ "+";
*/

static token_ * newToken(int type, double **variable, double constant,
		double *minDiv)
{
	token_ *r;
	r = malloc(sizeof(token_));
	ReturnNULLIf(r == NULL, "Malloc Failed");
	r->type = type;
	r->variable = variable;
	r->constant = constant;
	r->minDiv = minDiv;
	r->next = NULL;
	return r;
}

int tokenizerDestroyAll(token_ **r)
{
	token_ *ptr, *ptrNext;
	
	ReturnErrIf(r == NULL, "Bad Pointer");
	ReturnErrIf((*r) == NULL, "Not initialised");
	Debug("Destroying Tokenizer %p", (*r));
	
	for(ptr = (*r); ptr != NULL; ptr = ptrNext) {
		ptrNext = ptr->next;
		free(ptr);
	}
	
	*r = NULL;
	return 0;
}

#define Attach(tok, var, con, minDiv) \
	tail->next = newToken(tok, var, con, minDiv); \
	GotoFailedIf(tail->next == NULL); \
	tail = tail->next

token_ * tokenizerNew(token_ *r, char *buffer, tokenizerGetVarPtr getVarFunc, 
		void *private, double *minDiv)
{
	token_ *head = NULL, *tail = NULL;
	char *cursor = NULL, *marker = NULL, *start = NULL, tmp0, iv = 'v';
	double **variable;
	
	ReturnNULLIf(r != NULL);
	ReturnNULLIf(buffer == NULL);
	ReturnNULLIf(getVarFunc == NULL);
	
	Debug("Creating tokenizer");
	cursor = buffer;
	marker = buffer;
	head = newToken(-1, NULL, 0, NULL);
	ReturnNULLIf(head == NULL);
	tail = head;
	
parse:
	start = cursor;
/*!re2c /*!ignore!re2c Parse Block */
	(WSP+ | CONTINUE)	{ goto parse; }
	EOI 	{ Attach(0, NULL, 0, NULL);						goto passed; }
	'v('	{ Attach(TOKEN_LPAREN, NULL, 0, NULL); iv='v';	goto parse_iv; }
	'i('	{ Attach(TOKEN_LPAREN, NULL, 0, NULL); iv='i';	goto parse_iv; }
	'abs('	{ Attach(TOKEN_ABS, NULL, 0, minDiv);		goto parse; }
	'acosh('{ Attach(TOKEN_ACOSH, NULL, 0, minDiv);	goto parse; }
	'acos('	{ Attach(TOKEN_ACOS, NULL, 0, minDiv);		goto parse; }
	'asinh('{ Attach(TOKEN_ASINH, NULL, 0, minDiv);	goto parse; }
	'asin('	{ Attach(TOKEN_ASIN, NULL, 0, minDiv);		goto parse; }
	'atanh('{ Attach(TOKEN_ATANH, NULL, 0, minDiv);	goto parse; }
	'atan('	{ Attach(TOKEN_ATAN, NULL, 0, minDiv);		goto parse; }
	'cosh('	{ Attach(TOKEN_COSH, NULL, 0, minDiv);		goto parse; }
	'cos('	{ Attach(TOKEN_COS, NULL, 0, minDiv);		goto parse; }
	'exp('	{ Attach(TOKEN_EXP, NULL, 0, minDiv);		goto parse; }
	'ln('	{ Attach(TOKEN_LN, NULL, 0, minDiv);		goto parse; }
	'log('	{ Attach(TOKEN_LOG, NULL, 0, minDiv);		goto parse; }
	'sinh('	{ Attach(TOKEN_SINH, NULL, 0, minDiv);		goto parse; }
	'sin('	{ Attach(TOKEN_SIN, NULL, 0, minDiv);		goto parse; }
	'sqrt('	{ Attach(TOKEN_SQRT, NULL, 0, minDiv);		goto parse; }
	'tan('	{ Attach(TOKEN_TAN, NULL, 0, minDiv);		goto parse; }
	'uramp('{ Attach(TOKEN_URAMP, NULL, 0, minDiv);	goto parse; }
	'u('	{ Attach(TOKEN_U, NULL, 0, minDiv);		goto parse; }
	"+" 	{ Attach(TOKEN_PLUS, NULL, 0, minDiv);		goto parse; }
	"-" 	{ Attach(TOKEN_MINUS, NULL, 0, minDiv);	goto parse; }
	"*" 	{ Attach(TOKEN_TIMES, NULL, 0, minDiv);	goto parse; }
	"/" 	{ Attach(TOKEN_DIVIDE, NULL, 0, minDiv);	goto parse; }
	"^" 	{ Attach(TOKEN_POWER, NULL, 0, minDiv);	goto parse; }
	"(" 	{ Attach(TOKEN_LPAREN, NULL, 0, NULL);		goto parse; }
	")" 	{ Attach(TOKEN_RPAREN, NULL, 0, NULL);		goto parse; }
	NUMBER 	{ Attach(TOKEN_CONSTANT, NULL, atof(start), NULL); goto parse; }
	ALPHA+ 	{ 
		tmp0 = *cursor;
		*cursor = 0x0;
		variable = getVarFunc(start, 0x0, private);
		GotoFailedIf(variable == NULL, "User rejected the variable %s", start);
		Attach(TOKEN_VARIABLE, variable, 0, NULL);
		*cursor = tmp0;
		goto parse;
	}
	ANY 	{ Error("Unexpected character %s", cursor-1);	goto failed; }
*/

parse_iv:
	start = cursor;
/*!re2c /*!ignore!re2c Parse Block */
	(WSP+ | CONTINUE)	{ goto parse_iv; }
	")" 	{ Attach(TOKEN_RPAREN, NULL, 0, NULL);		goto parse; }
	','		{ Attach(TOKEN_MINUS, NULL, 0, NULL);		goto parse_iv; }
	(ALPHA | DIGIT)+ 	{ 
		tmp0 = *cursor;
		*cursor = 0x0;
		variable = getVarFunc(start, iv, private);
		GotoFailedIf(variable == NULL, "User rejected the variable %s", start);
		Attach(TOKEN_VARIABLE, variable, 0, NULL);
		*cursor = tmp0;
		goto parse_iv;
	}
	ANY 	{ cursor--; Error("Unexpected character %s", cursor);	goto failed; }
*/

passed:
	tail = head->next;
	free(head);
	return tail;
failed:
	tokenizerDestroyAll(&head);
	return NULL;
}
