/* 
 * Copyright (C) 2006 Cooper Street Innovations Inc.
 *	Charles Eidsness    <charles@cooper-street.com>
 *
 * This program is free software; you can redistribute it and/or
 * modify it under the terms of the GNU General Public License
 * as published by the Free Software Foundation; either version 2
 * of the License, or (at your option) any later version.
 * 
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 * 
 * You should have received a copy of the GNU General Public License
 * along with this program; if not, write to the Free Software
 * Foundation, Inc., 51 Franklin Street, Fifth Floor, Boston, MA 
 * 02110-1301, USA.
 *
 */

#include <math.h>

#include "log.h"
#include "deck_internal.h"

/* RE Format parser defines */
#define	YYCTYPE		char
#define	YYCURSOR	p->cursor
#define	YYMARKER	p->marker

#define NOT_A_SUBCKT 0
#define IS_A_SUBCKT 1

typedef struct {
	char *cursor;
	char *marker;
} parser_;

static int deckParser(deck_ *r, char *buffer, int subckt);

/*!re2c /*!ignore!re2c Set RE2C Options */
	re2c:yyfill:enable=0;
*/

/*!re2c /*!ignore!re2c Useful Regular Expressions */
	EOI = [\000];
	EOL = [\012] | [\015];
	ANY	= [\000-\377];
	ALPHA	= [a-zA-Z];
	DIGIT	= [0-9];
	TEXT = (ALPHA | DIGIT | [./_\\]);
	NOTEOL = [^\012\015\000] ;
	UNITS = TEXT*;
	WSP = (" ") | ([=(),]) | (EOL+ '+') | [\011];
	WSPNLP = (" ") | ([=),]) | (EOL+ '+') | [\011];
	WSPNRP = (" ") | ([=(,]) | (EOL+ '+') | [\011];
	COMMENTS = (EOL | ([*] .* EOL) | WSP+);
*/

static char * parseLine(parser_ *p)
{
	char *line;
	
	line = p->cursor;
/*!re2c /*!ignore!re2c Text Block Parse Block */
	(NOTEOL | (EOL+ '+'))+ EOL {
		*(p->cursor - 1) = 0x0;
		return line;
	}
	ANY {
		return NULL;
	}
*/
	ReturnNULL("Parser should never get here!");
}

static char * parseWord(parser_ *p)
{
	char *word;
	
	
parse_word:
	word = p->cursor;
/*!re2c /*!ignore!re2c Text Block Parse Block */
	TEXT+ (WSP | EOL) {
		*(p->cursor - 1) = 0x0;
		return word;
	}
	WSP { goto parse_word; }
	ANY {
		return NULL;
	}
*/
	ReturnNULL("Parser should never get here!");
}

static double parseNumber(parser_ *p)
{
	char *number;
	
parse_number:
	number = p->cursor;
/*!re2c /*!ignore!re2c Text Block Parse Block */
	((("-"|"+")? DIGIT+ ("." DIGIT+)?)|("." DIGIT+)) ('e' ("-"|"+")? DIGIT+)? 'T' UNITS {
		return (atof(number) * 1e12);
	}
	((("-"|"+")? DIGIT+ ("." DIGIT+)?)|("." DIGIT+)) ('e' ("-"|"+")? DIGIT+)? 'G' UNITS {
		return (atof(number) * 1e9);
	}
	((("-"|"+")? DIGIT+ ("." DIGIT+)?)|("." DIGIT+)) ('e' ("-"|"+")? DIGIT+)? 'k' UNITS {
		return (atof(number) * 1e3);
	}
	((("-"|"+")? DIGIT+ ("." DIGIT+)?)|("." DIGIT+)) ('e' ("-"|"+")? DIGIT+)? 'u' UNITS {
		return (atof(number) * 1e-6);
	}
	((("-"|"+")? DIGIT+ ("." DIGIT+)?)|("." DIGIT+)) ('e' ("-"|"+")? DIGIT+)? 'n' UNITS {
		return (atof(number) * 1e-9);
	}
	((("-"|"+")? DIGIT+ ("." DIGIT+)?)|("." DIGIT+)) ('e' ("-"|"+")? DIGIT+)? 'p' UNITS {
		return (atof(number) * 1e-12);
	}
	((("-"|"+")? DIGIT+ ("." DIGIT+)?)|("." DIGIT+)) ('e' ("-"|"+")? DIGIT+)? 'f' UNITS {
		return (atof(number) * 1e-15);
	}
	((("-"|"+")? DIGIT+ ("." DIGIT+)?)|("." DIGIT+)) ('e' ("-"|"+")? DIGIT+)? 'Meg' UNITS {
		return (atof(number) * 1e6);
	}
	((("-"|"+")? DIGIT+ ("." DIGIT+)?)|("." DIGIT+)) ('e' ("-"|"+")? DIGIT+)? 'mil' UNITS {
		return (atof(number) * 2.54e-5);
	}
	((("-"|"+")? DIGIT+ ("." DIGIT+)?)|("." DIGIT+)) ('e' ("-"|"+")? DIGIT+)? 'm' UNITS {
		return (atof(number) * 1e-3);
	}
	((("-"|"+")? DIGIT+ ("." DIGIT+)?)|("." DIGIT+)) ('e' ("-"|"+")? DIGIT+)? UNITS {
		return (atof(number));
	}
	WSP { goto parse_number; }
	ANY {
		p->cursor--;
		Warn("%s",p->cursor);
		return sqrtf(-1.f);
	}
*/
	ReturnErr("Parser should never get here!");
}

typedef struct {
	int rowType[DEVICE_MAX_PINS];
	int numPins;
} footPrint_;

footPrint_ ggFootPrint = { {TYPE_NORMAL, TYPE_NORMAL, TYPE_NONE, TYPE_NONE, }, 2 };
footPrint_ ggvFootPrint = { {TYPE_NORMAL, TYPE_NORMAL, TYPE_EXTENDED, TYPE_NONE, }, 3 };
footPrint_ ggggFootPrint = { {TYPE_NORMAL, TYPE_NORMAL, TYPE_NORMAL, TYPE_NORMAL, }, 4 };

static device_ * parseDevice(deck_ *r, parser_ *p, footPrint_ *fp, int subckt)
{
	char *refdes, *pinName;
	row_ *pin[DEVICE_MAX_PINS];
	device_ *newDevice = NULL;
	int i;
	
	refdes = parseWord(p);
	ReturnNULLIf(refdes == NULL);
	if(subckt == IS_A_SUBCKT) {
		refdes = subcktDBTransformRefDes(r->subckts, refdes);
		ReturnNULLIf(refdes == NULL);
	}
	for(i = 0; i < fp->numPins; i++) {
		pinName = parseWord(p);
		ReturnNULLIf(pinName == NULL, "pin %i %s", i, p->cursor);
		if(subckt == IS_A_SUBCKT) {
			pinName = subcktDBTransformNode(r->subckts, pinName);
			ReturnNULLIf(pinName == NULL, "pin %i", i);
		}
		pin[i] = matrixFindRow(r->matrix, pinName, fp->rowType[i]);
		ReturnNULLIf(pin[i] == NULL, "pin %i", i);
	}
	
	newDevice = deviceNew(newDevice, r->matrix, refdes, pin, fp->numPins);
	ReturnNULLIf(deviceDBAttach(r->devices, newDevice));
	return newDevice;
}

static int parseResistor(deck_ *r, parser_ *p, device_ *newDevice)
{
	double R;
	
	R = parseNumber(p);
	ReturnErrIf(isnan(R), "Failed to parse R");
	ReturnErrIf(deviceConfigResistor(newDevice, R));
	
	return 0;
}

static int parseInductor(deck_ *r, parser_ *p, device_ *newDevice)
{
	double L, ic = 0;
	
	L = parseNumber(p);
	ReturnErrIf(isnan(L), "Failed to parse L");
	
/*!re2c /*!ignore!re2c Text Block Parse Block */
	'IC' WSP+ ("-" | "+" | DIGIT) {
		p->cursor--;
		ic = parseNumber(p);
		ReturnErrIf(isnan(ic), "Failed to parse IC");
		goto done;
	}
	ANY {
		p->cursor--;
		goto done;
	}
*/
done:
	ReturnErrIf(deviceConfigInductor(newDevice, L, ic));
	
	return 0;
}

static int parseCapacitor(deck_ *r, parser_ *p, device_ *newDevice)
{
	double C, ic = 0;
	
	C = parseNumber(p);
	ReturnErrIf(isnan(C), "Failed to parse L");
	
/*!re2c /*!ignore!re2c Text Block Parse Block */
	'IC' WSP+ ("-" | "+" | DIGIT) {
		p->cursor--;
		ic = parseNumber(p);
		ReturnErrIf(isnan(ic), "Failed to parse IC");
		goto done;
	}
	ANY {
		p->cursor--;
		goto done;
	}
*/
done:
	ReturnErrIf(deviceConfigCapacitor(newDevice, C, ic));
	
	return 0;
}

static int countPWLs(parser_ *p)
{
	int pwlCount = 0;
count_pwls:
/*!re2c /*!ignore!re2c Text Block Parse Block */
	WSPNRP {
		goto count_pwls;
	}
	((("-" | "+")? DIGIT+ ("." DIGIT+)?) | ("." DIGIT+)) ('e' ("-" | "+")? DIGIT+)? UNITS {
		pwlCount++;
		goto count_pwls;
	}
	')' {
		return pwlCount;
	}
	ANY {
		ReturnErr("Bad pwl statement %s", p->cursor);
	}
*/
	ReturnErr("Parser should never get here!");
}

#define PULSE_LEN 7
#define SIN_LEN 5
#define EXP_LEN 6
#define SFFM_LEN 5
#define PWL_LEN 2
static int parseSource(deck_ *r, parser_ *p, device_ *newDevice, char type)
{
	double dc = 0;
	srcType_ srcType = SOURCE_NONE;
	int i, srcLength = 0;
	double *srcData = NULL;
	parser_ tmp;

parse_source:
/*!re2c /*!ignore!re2c Text Block Parse Block */
	(('DC' WSP+ ("-" | "+" | DIGIT | '.')) | ("-" | "+" | DIGIT | '.')) {
		p->cursor--;
		dc = parseNumber(p);
		ReturnErrIf(isnan(dc), "Failed to parse DC");
		goto parse_source;
	}
	'pulse(' {
		ReturnErrIf(srcType != SOURCE_NONE);
		srcType = SOURCE_PULSE;
		srcData = malloc(sizeof(double)*PULSE_LEN);
		srcData[0] = parseNumber(p);
		ReturnErrAndFreeIf(srcData, isnan(srcData[0]));
		srcLength++;
		srcData[1] = parseNumber(p);
		ReturnErrAndFreeIf(srcData, isnan(srcData[1]));
		srcLength++;
		for(i = 2; i < PULSE_LEN; i++) {
			srcData[i] = parseNumber(p);
			if(isnan(srcData[i])) {
				goto parse_source;
			}
			srcLength++;
		}
		goto parse_source;
	}
	'sin(' {
		ReturnErrIf(srcType != SOURCE_NONE);
		srcType = SOURCE_SIN;
		srcData = malloc(sizeof(double)*SIN_LEN);
		srcData[0] = parseNumber(p);
		ReturnErrAndFreeIf(srcData, isnan(srcData[0]));
		srcLength++;
		srcData[1] = parseNumber(p);
		ReturnErrAndFreeIf(srcData, isnan(srcData[1]));
		srcLength++;
		for(i = 2; i < SIN_LEN; i++) {
			srcData[i] = parseNumber(p);
			if(isnan(srcData[i])) {
				goto parse_source;
			}
			srcLength++;
		}
		goto parse_source;
	}
	'exp(' {
		ReturnErrIf(srcType != SOURCE_NONE);
		srcType = SOURCE_EXP;
		srcData = malloc(sizeof(double)*EXP_LEN);
		srcData[0] = parseNumber(p);
		ReturnErrAndFreeIf(srcData, isnan(srcData[0]));
		srcLength++;
		srcData[1] = parseNumber(p);
		ReturnErrAndFreeIf(srcData, isnan(srcData[1]));
		srcLength++;
		for(i = 2; i < EXP_LEN; i++) {
			srcData[i] = parseNumber(p);
			if(isnan(srcData[i])) {
				goto parse_source;
			}
			srcLength++;
		}
		goto parse_source;
	}
	'sffm(' {
		ReturnErrIf(srcType != SOURCE_NONE);
		srcType = SOURCE_SFFM;
		srcData = malloc(sizeof(double)*SFFM_LEN);
		srcData[0] = parseNumber(p);
		ReturnErrAndFreeIf(srcData, isnan(srcData[0]));
		srcLength++;
		srcData[1] = parseNumber(p);
		ReturnErrAndFreeIf(srcData, isnan(srcData[1]));
		srcLength++;
		for(i = 2; i < SFFM_LEN; i++) {
			srcData[i] = parseNumber(p);
			if(isnan(srcData[i])) {
				goto parse_source;
			}
			srcLength++;
		}
		goto parse_source;
	}
	'pwl(' {
		ReturnErrIf(srcType != SOURCE_NONE);
		srcType = SOURCE_PWL;
		tmp.cursor = p->cursor;
		tmp.marker = p->marker;
		srcLength = countPWLs(&tmp);
		ReturnErrIf(srcLength < 0);
		ReturnErrIf(MOD(srcLength, PWL_LEN) != 0);
		srcData = malloc(sizeof(double)*srcLength);
		for(i = 0; i < srcLength; i++) {
			srcData[i] = parseNumber(p);
			ReturnErrAndFreeIf(srcData, isnan(srcData[i]));
		}
		goto parse_source;
	}
	WSP { goto parse_source; }
	EOL { goto done; }
	ANY {
		p->cursor--;
		if(srcData != NULL) {
			free(srcData);
		}
		ReturnErr("Expeciting DC value, pulse, sin, sffm, pwl, or exp");
	}
*/
done:
	ReturnErrIf(deviceConfigSource(newDevice, dc, srcType, srcData,
			srcLength, type));
	
	return 0;
}

static int parseVCCS(deck_ *r, parser_ *p, device_ *newDevice)
{
	double gain;
	
	gain = parseNumber(p);
	ReturnErrIf(isnan(gain), "Failed to parse gain");
	ReturnErrIf(deviceConfigVCCS(newDevice, gain));
	
	return 0;
}

static int parseVCVS(deck_ *r, parser_ *p, device_ *newDevice)
{
	double gain;
	
	gain = parseNumber(p);
	ReturnErrIf(isnan(gain), "Failed to parse gain");
	ReturnErrIf(deviceConfigVCVS(newDevice, gain));
	
	return 0;
}

static int parseCCCS(deck_ *r, parser_ *p, device_ *newDevice)
{
	double gain;
	
	gain = parseNumber(p);
	ReturnErrIf(isnan(gain), "Failed to parse gain");
	ReturnErrIf(deviceConfigCCCS(newDevice, gain));
	
	return 0;
}

static int parseCCVS(deck_ *r, parser_ *p, device_ *newDevice)
{
	double gain;
	
	gain = parseNumber(p);
	ReturnErrIf(isnan(gain), "Failed to parse gain");
	ReturnErrIf(deviceConfigCCVS(newDevice, gain));
	
	return 0;
}

static int parseNonlinear(deck_ *r, parser_ *p, device_ *newDevice)
{
	char *equation;
	char type;
	
/*!re2c /*!ignore!re2c Text Block Parse Block */
	'v' WSP* '=' WSPNLP* {
		type = 'v';
		goto get_equation;
	}
	'i' WSP* '=' WSPNLP* {
		type = 'i';
		goto get_equation;
	}
	ANY {
		p->cursor--;
		ReturnErr("B element must be either \"v=\" or \"i=\"");
	}
*/

get_equation:
	equation = parseLine(p);
	ReturnErrIf((equation == 0), "Failed to parse equation");
	if(type == 'v') {
		ReturnErrIf(deviceConfigNonlinearV(newDevice, equation));
	} else if(type == 'i') {
		ReturnErrIf(deviceConfigNonlinearI(newDevice, equation));
	} else {
		ReturnErr("Don't support non-linear type %c", type);
	}
	
	return 0;
}

static int parseTLine(deck_ *r, parser_ *p, device_ *newDevice)
{
	double z0 = 0, td = 0, f, nl = 0.25;

parse_tline:
/*!re2c /*!ignore!re2c Text Block Parse Block */
	(('Z0' WSP+ ("-" | "+" | DIGIT)) | ("-" | "+" | DIGIT)) {
		p->cursor--;
		z0 = parseNumber(p);
		ReturnErrIf(isnan(z0), "Failed to parse Z0");
		goto parse_tline;
	}
	(('TD' WSP+ ("-" | "+" | DIGIT)) | ("-" | "+" | DIGIT)) {
		p->cursor--;
		td = parseNumber(p);
		ReturnErrIf(isnan(td), "Failed to parse Td");
		goto parse_tline;
	}
	(('F' WSP+ ("-" | "+" | DIGIT)) | ("-" | "+" | DIGIT)) {
		p->cursor--;
		f = parseNumber(p);
		ReturnErrIf(isnan(f), "Failed to parse F");
		ReturnErrIf(f <= 0, "Frequency must be > 0");
		nl = parseNumber(p);
		if(isnan(nl)) {
			nl = 0.25;
		}
		td = nl/f;
		goto parse_tline;
	}
	WSP { goto parse_tline; }
	EOL { goto done; }
	ANY {
		p->cursor--;
		ReturnErr("Expeciting Z0, TD, or F");
	}
*/
done:
	ReturnErrIf(z0 == 0, "Must specifiy a transmission line impedance");
	ReturnErrIf(td == 0, "Must specifiy a transmission line length");
	ReturnErrIf(deviceConfigTLine(newDevice, z0, td));
	
	return 0;
}

static int countWords(parser_ *p)
{
	int wordCount = 0;
count_words:
/*!re2c /*!ignore!re2c Text Block Parse Block */
	TEXT+ {
		wordCount++;
		goto count_words;
	}
	WSP {
		goto count_words;
	}
	ANY {
		return wordCount;
	}
*/
	ReturnErr("Parser should never get here!");
}

static int parseSubcktInstance(deck_ *r, parser_ *p, int subckt)
{
	char *refdes, **nodes, *name;
	parser_ tmp;
	int i, numNodes;
	
	refdes = parseWord(p);
	ReturnErrIf(refdes == NULL);
	if(subckt == IS_A_SUBCKT) {
		refdes = subcktDBTransformRefDes(r->subckts, refdes);
		ReturnErrIf(refdes == NULL);
	}
	tmp.cursor = p->cursor;
	tmp.marker = p->marker;
	numNodes = countWords(&tmp) - 1;
	ReturnErrIf(numNodes < 0);
	nodes = malloc(sizeof(char*)*numNodes);
	ReturnErrIf(nodes == NULL);
	for(i = 0; i < numNodes; i++) {
		nodes[i] = parseWord(p);
		ReturnErrIf(nodes[i] == NULL, "pin %i", i);
		if(subckt == IS_A_SUBCKT) {
			nodes[i] = subcktDBTransformNode(r->subckts, nodes[i]);
			ReturnErrIf(nodes[i] == NULL, "pin %i", i);
		}
	}
	name = parseWord(p);
	ReturnErrIf(name == NULL);
	
	ReturnErrIf(subcktDBAddXElement(r->subckts, refdes, nodes, numNodes, name));
	
	return 0;
}

static int parseInclude(deck_ *r, parser_ *p, int subckt)
{
	char *filename, *buffer;
	
	filename = parseWord(p);
	ReturnErrIf(filename == NULL);
	ReturnErrIf(!strcmp(filename,r->filename), "Recursive include");
	
	buffer = includeDBAddFile(r->includes, filename);
	ReturnErrIf(buffer == NULL);
	
	ReturnErrIf(deckParser(r, buffer, subckt));
	
	return 0;
}

static int parseSubckt(deck_ *r, parser_ *p)
{
	char *name, **nodes, *buffer;
	parser_ tmp;
	int i, numNodes;
	
	name = parseWord(p);
	ReturnErrIf(name == NULL);
	tmp.cursor = p->cursor;
	tmp.marker = p->marker;
	numNodes = countWords(&tmp);
	ReturnErrIf(numNodes < 0);
	nodes = malloc(sizeof(char*)*(numNodes));
	ReturnErrIf(nodes == NULL);
	for(i = 0; i < numNodes; i++) {
		nodes[i] = parseWord(p);
		ReturnErrIf(nodes[i] == NULL, "pin %i", i);
	}
	buffer = p->cursor;
	
parse_subckt:
/*!re2c /*!ignore!re2c Text Block Parse Block */
	'.ends' NOTEOL* {
		*(p->cursor - 1) = 0x0;
		ReturnErrIf(subcktDBAddSubckt(r->subckts, name, nodes, numNodes,
				buffer));
		return 0;
	}
	EOI {
		ReturnErr("Couldn't find .ends");
	}
	ANY {
		goto parse_subckt;
	}
*/
	ReturnErr("Parser should never get here!");
}

static int parseOpPoint(deck_ *r, parser_ *p)
{
	command_ *newCommand = NULL;
	
	newCommand = commandNew(newCommand, r->matrix, r->devices);
	ReturnErrIf(newCommand == NULL);
	ReturnErrIf(commandDBAttach(r->commands, newCommand));
	ReturnErrIf(commandConfigOpPoint(newCommand));
	
	return 0;
}

static int parseTransient(deck_ *r, parser_ *p)
{
	command_ *newCommand = NULL;
	double tstep, tstop, tstart, tmax;
	int uic = 0;
	
	tstart = sqrtf(-1.f);
	tmax = sqrtf(-1.f);
	
	tstep = parseNumber(p);
	ReturnErrIf(isnan(tstep), "Failed to parse tstep");
	tstop = parseNumber(p);
	ReturnErrIf(isnan(tstop), "Failed to parse tstop");
	
parse_state:
/*!re2c /*!ignore!re2c Text Block Parse Block */
	WSP* ("-" | "+" | DIGIT) WSP* {
		p->cursor--;
		tstart = parseNumber(p);
		ReturnErrIf(isnan(tstart), "Failed to parse tstart");
		tmax = parseNumber(p);
		goto parse_state;
	}
	WSP* 'uic' WSP* {
		uic = 1;
		goto parse_state;
	}
	ANY {
		p->cursor--;
		goto create_command;
	}
*/

create_command:
	
	newCommand = commandNew(newCommand, r->matrix, r->devices);
	ReturnErrIf(newCommand == NULL);
	ReturnErrIf(commandDBAttach(r->commands, newCommand));
	ReturnErrIf(commandConfigTransient(newCommand, tstep, tstop, tstart, 
			tmax, uic));
	
	return 0;
}

static int deckParser(deck_ *r, char *buffer, int subckt)
{
	int line = 1;
	device_ *newDevice;
	parser_ _p, *p;
	p = &_p;
	
	p->cursor = buffer;
	p->marker = buffer;
	
	/* Only have the title line on the primary file */
	if(buffer != r->buffer)
		goto parse;

deck_name:
/*!re2c /*!ignore!re2c Parse Block */
	EOL {
		*(p->cursor - 1) = 0x0;
		goto parse;
	}
	EOI {
		ReturnErr("No data in file");
	}
	ANY {
		goto deck_name;
	}
*/

parse:
	line++;
/*!re2c /*!ignore!re2c Parse Block */
	COMMENTS {
		*(p->cursor - 1) = 0x0;
		goto parse;
	}
	'R' {
		p->cursor--;
		newDevice = parseDevice(r, p, &ggFootPrint, subckt);
		ReturnErrIf(newDevice == NULL);
		ReturnErrIf(parseResistor(r, p, newDevice));
		goto parse;
	}
	'L' {
		p->cursor--;
		newDevice = parseDevice(r, p, &ggFootPrint, subckt);
		ReturnErrIf(newDevice == NULL);
		ReturnErrIf(parseInductor(r, p, newDevice));
		goto parse;
	}
	'C' {
		p->cursor--;
		newDevice = parseDevice(r, p, &ggFootPrint, subckt);
		ReturnErrIf(newDevice == NULL);
		ReturnErrIf(parseCapacitor(r, p, newDevice));
		goto parse;
	}
	'V' {
		p->cursor--;
		newDevice = parseDevice(r, p, &ggFootPrint, subckt);
		ReturnErrIf(newDevice == NULL);
		ReturnErrIf(parseSource(r, p, newDevice, 'v'));
		goto parse;
	}
	'I' {
		p->cursor--;
		newDevice = parseDevice(r, p, &ggFootPrint, subckt);
		ReturnErrIf(newDevice == NULL);
		ReturnErrIf(parseSource(r, p, newDevice, 'i'));
		goto parse;
	}
	'G' {
		p->cursor--;
		newDevice = parseDevice(r, p, &ggggFootPrint, subckt);
		ReturnErrIf(newDevice == NULL);
		ReturnErrIf(parseVCCS(r, p, newDevice));
		goto parse;
	}
	'E' {
		p->cursor--;
		newDevice = parseDevice(r, p, &ggggFootPrint, subckt);
		ReturnErrIf(newDevice == NULL);
		ReturnErrIf(parseVCVS(r, p, newDevice));
		goto parse;
	}
	'F' {
		p->cursor--;
		newDevice = parseDevice(r, p, &ggvFootPrint, subckt);
		ReturnErrIf(newDevice == NULL);
		ReturnErrIf(parseCCCS(r, p, newDevice));
		goto parse;
	}
	'H' {
		p->cursor--;
		newDevice = parseDevice(r, p, &ggvFootPrint, subckt);
		ReturnErrIf(newDevice == NULL);
		ReturnErrIf(parseCCVS(r, p, newDevice));
		goto parse;
	}
	'B' {
		p->cursor--;
		newDevice = parseDevice(r, p, &ggFootPrint, subckt);
		ReturnErrIf(newDevice == NULL);
		ReturnErrIf(parseNonlinear(r, p, newDevice));
		goto parse;
	}
	'T' {
		p->cursor--;
		newDevice = parseDevice(r, p, &ggggFootPrint, subckt);
		ReturnErrIf(newDevice == NULL);
		ReturnErrIf(parseTLine(r, p, newDevice));
		goto parse;
	}
	'X' {
		p->cursor--;
		ReturnErrIf(parseSubcktInstance(r, p, subckt));
		goto parse;
	}
	'.include' WSP+ {
		ReturnErrIf(parseInclude(r, p, subckt));
		goto parse;
	}
	'.subckt' WSP+ {
		ReturnErrIf(parseSubckt(r, p));
		goto parse;
	}
	'.op' WSP* EOL {
		ReturnErrIf(parseOpPoint(r,p));
		goto parse;
	}
	'.tran' WSP+ {
		ReturnErrIf(parseTransient(r,p));
		goto parse;
	}
	'.end' 's'? {
		return 0;
	}
	EOI {
		if(buffer == r->buffer) {
			Warn("File missing .end line");
		}
		return 0;
	}
	ANY {
		ReturnErr("Unsuported device on line %i %.256s", line, p->cursor-1);
	}
*/
	
	ReturnErr("Parser should never get here!");
}

int deckParse(deck_ *r)
{
	subcktXListStatus subcktStatus;
	char *buffer;
	
	ReturnErrIf(r == NULL);
	
	ReturnErrIf(deckParser(r, r->buffer, NOT_A_SUBCKT));
	
	subcktStatus = subcktDBNextX(r->subckts);
	while(subcktStatus == X_VALID) {
		buffer = subcktDBGetActiveBuffer(r->subckts);
		ReturnErrIf(buffer == NULL);
		ReturnErrIf(deckParser(r, buffer, IS_A_SUBCKT));
		subcktStatus = subcktDBNextX(r->subckts);
	}
	ReturnErrIf(subcktStatus == X_ERR);
	
	return 0;
}
