/* 
 * Copyright (C) 2006 Cooper Street Innovations Inc.
 *	Charles Eidsness    <charles@cooper-street.com>
 *
 * This program is free software; you can redistribute it and/or
 * modify it under the terms of the GNU General Public License
 * as published by the Free Software Foundation; either version 2
 * of the License, or (at your option) any later version.
 * 
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 * 
 * You should have received a copy of the GNU General Public License
 * along with this program; if not, write to the Free Software
 * Foundation, Inc., 51 Franklin Street, Fifth Floor, Boston, MA 
 * 02110-1301, USA.
 *
 */

#include <log.h>
#include "matrix_internal.h"

int matrixReSolve(matrix_ *r)
{
	ReturnErrIf(r == NULL);
	ReturnErrIf(r->reSolve == NULL);
	ReturnErrIf(r->reSolve(r));
	return 0;
}

int matrixSolve(matrix_ *r)
{
	ReturnErrIf(r == NULL);
	ReturnErrIf(r->solve == NULL);
	ReturnErrIf(r->solve(r));
	return 0;
}

int matrixNumRows(matrix_ *r)
{
	int length;
	ReturnErrIf(r == NULL);
	length = rowDBLength(r->rows);
	ReturnErrIf(length < 0);
	return length;
}

int matrixNumTimesteps(matrix_ *r)
{
	int length;
	ReturnErrIf(r == NULL);
	length = resultDBLength(r->results);
	ReturnErrIf(length < 0);
	return length;
}

resultDB_ * matrixGetResults(matrix_ *r)
{
	ReturnNULLIf(r == NULL);
	return r->results;
}

row_ * matrixFindRow(matrix_ *r, char *name, rowType_ type)
{
	ReturnNULLIf(r == NULL);
	return rowDBFind(r->rows, name, type);
}

node_ * matrixFindNode(matrix_ *r, row_ *row, row_ *col)
{
	ReturnNULLIf(r == NULL);
	return nodeDBFind(r->nodeDB, rowGetIndex(row), rowGetIndex(col));
}

int matrixWriteRaw(matrix_ *r, FILE *outFile)
{
	int n;
	
	ReturnErrIf(r == NULL);
	ReturnErrIf(outFile == NULL);
	
	n = rowDBLength(r->rows);
	ReturnErrIf(n < 0);
	fprintf(outFile, "No. Variables: %i\n", n + 1);
	n = resultDBLength(r->results);
	ReturnErrIf(n < 0);
	fprintf(outFile, "No. Points: %i\n", n);
	fprintf(outFile, "Variables:\n");
	fprintf(outFile, "\t0\ttime\ttime\n");
	ReturnErrIf(rowDBWriteRaw(r->rows, outFile));
	ReturnErrIf(resultDBWriteRaw(r->results, outFile));
	
	return 0;
}

int matrixDumpData(matrix_ *r)
{
	int i, j, k;
	
	ReturnErrIf(r == NULL);
	
	Debug("Matrix A^T:");
	for(i = 0, k = 0; i < (r->lenXB); i++) {
		for(j = 0; j < (r->lenXB); j++) {
			if(r->aRow[k] == j && (k < r->lenA)) {
				Text("%e\t", r->A[k]);
				k++;
			} else {
				Text("%e\t", 0.0);
			}
		}
		Text("\n");
	}
	Debug("Matrix B:");
	for(i = 0; i < (r->lenXB); i++) {
		Text("\t%i:%e\n", i, r->B[i]);
	}
	Debug("Matrix X:");
	for(i = 0; i < (r->lenXB); i++) {
		Text("\t%i:%e\n", i, r->X[i]);
	}
	
	return 0;
}

int matrixRecall(matrix_ *r)
{
	ReturnErrIf(r == NULL);
	ReturnErrIf(resultDBRecall(r->results, r->X, r->lenXB));
	return 0;
}

int matrixRecord(matrix_ *r, double time, int brkPoint)
{
	ReturnErrIf(r == NULL);
	ReturnErrIf(resultDBRecord(r->results, time, r->X, r->lenXB, brkPoint));
	return 0;
}

int matrixInit(matrix_ *r, state_ *state)
{
	ReturnErrIf(r == NULL);
	
	r->lenA = nodeDBLength(r->nodeDB);
	ReturnErrIf(r->lenA < 0);
	r->lenXB = rowDBLength(r->rows);
	ReturnErrIf(r->lenXB < 0);
	
	r->A = calloc(r->lenA, sizeof(double));
	ReturnErrIf(r->A == NULL);
	
	r->aRow = calloc(r->lenA, sizeof(int));
	ReturnErrIf(r->aRow == NULL);
	
	r->aColStart = calloc((r->lenXB+1), sizeof(int));
	ReturnErrIf(r->aColStart == NULL);
	
	r->X = calloc(r->lenXB, sizeof(double));
	ReturnErrIf(r->X == NULL);
	
	r->B = calloc(r->lenXB, sizeof(double));
	ReturnErrIf(r->B == NULL);
	
	ReturnErrIf(nodeDBInit(r->nodeDB, r->A, r->aRow, r->aColStart));
	ReturnErrIf(rowDBInit(r->rows, r->B, r->X));
	
	switch(stateGetLULibrary(state)) {
	case LU_ERR:
		ReturnErr("Unrecognized LU library");
	case LU_SUPERLU:
		ReturnErrIf(matrixInitSuperLU(r));
		break;
	}
	
	return 0;
}

int matrixDestroy(matrix_ **r)
{
	ReturnErrIf(r == NULL);
	ReturnErrIf((*r) == NULL);
	
	Debug("Destroying Matrix %p", (*r));
	
	ReturnErrIf(nodeDBDestroy(&(*r)->nodeDB));
	ReturnErrIf(rowDBDestroy(&(*r)->rows));
	ReturnErrIf(resultDBDestroy(&(*r)->results));
	
	if((*r)->unconfig != NULL) {
		ReturnErrIf((*r)->unconfig(*r));
	}
	
	if((*r)->A != NULL)
		free((*r)->A);
	if((*r)->aRow != NULL)
		free((*r)->aRow);
	if((*r)->aColStart != NULL)
		free((*r)->aColStart);
	if((*r)->X != NULL)
		free((*r)->X);
	if((*r)->B != NULL)
		free((*r)->B);
	
	if((*r)->library != NULL)
		free((*r)->library);
	
	free(*r);
	*r = NULL;
	return 0;
}

static char * defaultName = "None";

matrix_ * matrixNew(matrix_ *r)
{
	ReturnNULLIf(r != NULL);
	r = calloc(1, sizeof(matrix_));
	ReturnNULLIf(r == NULL);
	
	Debug("Creating Matrix %p", r);
	
	/* Create gnd matrix and rows */
	r->rows = rowDBNew(r->rows);
	ReturnNULLAndFreeIf(r, r->rows == NULL);
	r->nodeDB = nodeDBNew(r->nodeDB);
	ReturnNULLAndFreeIf(r, r->nodeDB == NULL);
	r->results = resultDBNew(r->results);
	ReturnNULLAndFreeIf(r, r->results == NULL);
	
	r->name = defaultName;
	
	return r;
}

