/* 
 * Copyright (C) 2006 Cooper Street Innovations Inc.
 *	Charles Eidsness    <charles@cooper-street.com>
 *
 * This program is free software; you can redistribute it and/or
 * modify it under the terms of the GNU General Public License
 * as published by the Free Software Foundation; either version 2
 * of the License, or (at your option) any later version.
 * 
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 * 
 * You should have received a copy of the GNU General Public License
 * along with this program; if not, write to the Free Software
 * Foundation, Inc., 51 Franklin Street, Fifth Floor, Boston, MA 
 * 02110-1301, USA.
 *
 */

#include <math.h>
#include <log.h>
#include "row.h"
#include "result.h"

/*****************************************************************************/

struct _resultDB {
	double time;
	double *data;
	int length;
	int brkPoint;
	resultDB_ *next;
	resultDB_ *prev;
};

int resultDBWriteRaw(resultDB_ *r, FILE *outFile)
{
	resultDB_ *ptr;
	int length;
	
	ReturnErrIf(r == NULL);
	ReturnErrIf(outFile == NULL);
	
	fprintf(outFile, "Binary:\n");
	for(ptr = r; ptr != NULL; ptr = ptr->next) {
		if(ptr->data != NULL) {
			length = fwrite(&ptr->time, sizeof(double), 1, outFile);
			ReturnErrIf(length != 1);
			length = fwrite(ptr->data, sizeof(double), ptr->length, outFile);
			ReturnErrIf(length != ptr->length);
		}
	}
	
	return 0;
}

double resultDBPrevTime(resultDB_ *r)
{
	ReturnNaNIf(r == NULL);
	return r->prev->time;
}

resultDB_ * resultDBFindTime(resultDB_ *r, double time)
{
	ReturnNULLIf(r == NULL);
	ReturnNULLIf(time < 0.0);
	
	while(r != NULL) {
		if(r->time <= time) {
			if(r->next == NULL) {
				return r;
			} else if(r->next->time > time) {
				return r;
			}
			r = r->next;
		} else {
			ReturnNULLIf(r->prev == NULL);
			if(r->prev->next == NULL) {
				Mark;
				return r;
			}
			r = r->prev;
		}
	}
	
	ReturnNULL("Shouldn't get here");
}

resultDB_ * resultDBNextBreak(resultDB_ *r, double time, double *nextBreak)
{
	ReturnNULLIf(r == NULL);
	ReturnNULLIf(time < 0.0);
	ReturnNULLIf(nextBreak == NULL);
	
	r = resultDBFindTime(r, time);
	ReturnNULLIf(r == NULL);
	
	for(; r->next != NULL; r = r->next) {
		if(r->brkPoint) {
			*nextBreak = r->time;
			return r;
		}
	}
	
	*nextBreak = -1.0;
	return r;
}

resultDB_ * resultDBLinterp(resultDB_ *r, double time, int index, double *data)
{
	double t0, t1, x0, x1;
	
	ReturnNULLIf(r == NULL);
	ReturnNULLIf(time < 0.0);
	ReturnNULLIf(index < 0);
	ReturnNULLIf(data == NULL);
	
	if(index == rowGetIndex(&gndRow)) {
		*data = rowGetSolution(&gndRow);
		return r;
	}
	
	r = resultDBFindTime(r, time);
	ReturnNULLIf(r == NULL);
	ReturnNULLIf(r->data == NULL);
	
	if(r->next == NULL) {
		ReturnNULLIf(r->prev == NULL);
		ReturnNULLIf(r->prev->data == NULL);
		x0 = r->prev->data[index - 1];
		x1 = r->data[index - 1];
		t0 = r->prev->time;
		t1 = r->time;
	} else {
		ReturnNULLIf(r->next->data == NULL);
		x0 = r->data[index - 1];
		x1 = r->next->data[index - 1];
		t0 = r->time;
		t1 = r->next->time;
	}
	
	ReturnNULLIf((t1-t0) == 0);
	
	*data = ((x1-x0)/(t1-t0))*(time-t0) + x0;
	
	return r;
}

/* TODO: This could be made more efficient */
int resultDBLength(resultDB_ *r)
{
	int i;
	resultDB_ *ptr;
	ReturnErrIf(r == NULL);
	for(ptr = r, i = 0; ptr->next != NULL; i++, ptr = ptr->next);
	return i;
}

int resultDBRecall(resultDB_ *r, double *data, int length)
{
	ReturnErrIf(r == NULL);
	ReturnErrIf(r->prev == NULL);
	ReturnErrIf(r->prev->data == NULL);
	ReturnErrIf(length != r->prev->length);
	
	memcpy(data, r->prev->data, r->prev->length*sizeof(double));
	
	return 0;
}

int resultDBRecord(resultDB_ *r, double time, double *data, int length, int brkPoint)
{
	resultDB_ *newElement;
	
	ReturnErrIf(r == NULL);
	ReturnErrIf(data == NULL);
	ReturnErrIf(length < 1);
	ReturnErrIf(time < r->prev->time);
	
	newElement = calloc(1, sizeof(resultDB_));
	ReturnErrIf(newElement == NULL);
	newElement->length = length;
	newElement->brkPoint = brkPoint;
	newElement->time = time;
	newElement->data = malloc(length*sizeof(double));
	ReturnErrIf(newElement->data == NULL);
	memcpy(newElement->data, data, length*sizeof(double));
	
	newElement->next = NULL;
	newElement->prev = r->prev;
	r->prev->next = newElement;
	r->prev = newElement;
	
	return 0;
}

int resultDBDestroy(resultDB_ **r)
{
	resultDB_ *ptr, *ptrNext;
	
	ReturnErrIf(r == NULL);
	ReturnErrIf((*r) == NULL);
	Debug("Destroying result DB %p", (*r));
	
	for(ptr = (*r); ptr != NULL; ptr = ptrNext) {
		ptrNext = ptr->next;
		if(ptr->data != NULL) {
			free(ptr->data);
		}
		free(ptr);
	};
	
	return 0;
}

resultDB_ * resultDBNew(resultDB_ *r)
{
	ReturnNULLIf(r != NULL);
	r = calloc(1, sizeof(resultDB_));
	ReturnNULLIf(r == NULL);
	
	Debug("Creating Result DB %p", r);
	r->next = NULL;
	r->prev = r;
	
	return r;
}

