/* 
 * Copyright (C) 2006 Cooper Street Innovations Inc.
 *	Charles Eidsness    <charles@cooper-street.com>
 *
 * This program is free software; you can redistribute it and/or
 * modify it under the terms of the GNU General Public License
 * as published by the Free Software Foundation; either version 2
 * of the License, or (at your option) any later version.
 * 
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 * 
 * You should have received a copy of the GNU General Public License
 * along with this program; if not, write to the Free Software
 * Foundation, Inc., 51 Franklin Street, Fifth Floor, Boston, MA 
 * 02110-1301, USA.
 *
 */

#include <math.h>
#include <log.h>
#include "state.h"

#define Min2(x,y) ((x > y) ? y : x)
#define Max2(x,y) ((x > y) ? x : y)

struct _state {
	/* Old Spice Options */
	int itl1;
	int itl4;
	double reltol;
	double vntol;
	double abstol;
	double chgtol;
	double trtol;
	double minbreak;
	double gmin;
	
	/* New Eispice Options */
	stateLULibrary luLibrary;
	
	/* Transient Analysis State */
	double tstop;
	double tstep;
};

/*****************************************************************************/

double stateGetAbsError(state_ *r, double i)
{
	ReturnNaNIf(r == NULL);
	return r->reltol*i + r->abstol;
}

double stateGetVnError(state_ *r, double v)
{
	ReturnNaNIf(r == NULL);
	return r->reltol*v + r->vntol;
}

double stateGetChgError(state_ *r, double ch)
{
	ReturnNaNIf(r == NULL);
	return r->reltol * Max2(ch, r->chgtol);
}

double stateGetTrAbsError(state_ *r, double dd, double e)
{
	ReturnNaNIf(r == NULL);
	return r->trtol * e / Max2(dd , r->abstol);
}

double stateGetTrVnError(state_ *r, double dd, double e)
{
	ReturnNaNIf(r == NULL);
	return r->trtol * e / Max2(dd , r->vntol);
}

/*****************************************************************************/

int stateGetITL1(state_ *r)
{
	ReturnErrIf(r == NULL);
	return r->itl1;
}

int stateGetITL4(state_ *r)
{
	ReturnErrIf(r == NULL);
	return r->itl4;
}

double stateGetMinbreak(state_ *r)
{
	ReturnNaNIf(r == NULL);
	return r->minbreak;
}

double stateGetTStop(state_ *r)
{
	ReturnNaNIf(r == NULL);
	return r->tstop;
}

double stateGetTStep(state_ *r)
{
	ReturnNaNIf(r == NULL);
	return r->tstep;
}

stateLULibrary stateGetLULibrary(state_ *r)
{
	ReturnErrIf(r == NULL);
	return r->luLibrary;
}

double stateGetGmin(state_ *r)
{
	ReturnNaNIf(r == NULL);
	return r->gmin;
}

/*****************************************************************************/

int stateSetMinbreak(state_ *r, double minbreak)
{ 
	ReturnErrIf(r == NULL);
	ReturnErrIf(minbreak < 0);
	r->minbreak = minbreak ;
	return 0;
}

int stateSetTStop(state_ *r, double tstop)
{
	ReturnErrIf(r == NULL);
	ReturnErrIf(tstop < 0);
	r->tstop = tstop;
	return 0;
}

int stateSetTStep(state_ *r, double tstep)
{
	ReturnErrIf(r == NULL);
	ReturnErrIf(tstep < 0);
	r->tstep = tstep;
	return 0;
}

/*****************************************************************************/

int stateMinbreakIsSet(state_ *r)
{
	ReturnErrIf(r == NULL);
	return (r->minbreak < 0) ? 0 : 1;
}

/*****************************************************************************/

int stateDestroy(state_ **r)
{
	ReturnErrIf(r == NULL);
	ReturnErrIf(*r == NULL);
	Debug("Destroying Options %p", *r);
	
	free(*r);
	*r = NULL;
	return 0;
}

state_ * stateNew(state_ *r)
{
	ReturnNULLIf(r != NULL);
	r = calloc(1, sizeof(state_));
	ReturnNULLIf(r == NULL);
	
	Debug("Creating Options %p", r);
	
	r->itl1 = 100;
	r->itl4 = 10;
	r->reltol = 0.001;
	r->vntol = 1e-6;
	r->abstol = 1e-12;
	r->chgtol = 1e-14;
	r->trtol = 7;
	r->minbreak = -1.0;
	r->gmin = 1e-12;
	
	r->luLibrary = LU_SUPERLU;
	
	return r;
}
