/* 
 * Copyright (C) 2006 Cooper Street Innovations Inc.
 *	Charles Eidsness    <charles@cooper-street.com>
 *
 * This program is free software; you can redistribute it and/or
 * modify it under the terms of the GNU General Public License
 * as published by the Free Software Foundation; either version 2
 * of the License, or (at your option) any later version.
 * 
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 * 
 * You should have received a copy of the GNU General Public License
 * along with this program; if not, write to the Free Software
 * Foundation, Inc., 51 Franklin Street, Fifth Floor, Boston, MA 
 * 02110-1301, USA.
 *
 */

#include <math.h>
#include <log.h>
#include "device_internal.h"

static char *typeV = "Independant Voltage Source";
static char *typeI = "Independant Current Source";

#define NUM_PINS	2
#define K			0
#define J			1

struct _devicePrivate {
	double dc;
	srcType_ srcType;
	double *srcData;
	int srcLength;
	int pwlCount;
	node_ *nodeRK;
	node_ *nodeRJ;
	node_ *nodeKR;
	node_ *nodeJR;
	row_ *rowR;
	row_ *rowK;
	row_ *rowJ;
};

enum {
	PULSE_V1 = 0,
	PULSE_V2 = 1,
	PULSE_TD = 2,
	PULSE_TR = 3,
	PULSE_TF = 4,
	PULSE_PW = 5,
	PULSE_PER = 6,
	PULSE_LEN = 7
};

enum {
	SIN_VO = 0,
	SIN_VA = 1,
	SIN_FREQ = 2,
	SIN_TD = 3,
	SIN_THETA = 4, 
	SIN_LEN = 5
};

enum {
	EXP_V1 = 0,
	EXP_V2 = 1,
	EXP_TD1 = 2,
	EXP_TAU1 = 3,
	EXP_TD2 = 4,
	EXP_TAU2 = 5,
	EXP_LEN = 6
};

enum {
	PWL_T = 0,
	PWL_V = 1,
	PWL_LEN = 2
};

enum {
	SFFM_VO = 0,
	SFFM_VA = 1,
	SFFM_FC = 2,
	SFFM_MDI = 3,
	SFFM_FS = 4,
	SFFM_LEN = 5
};

static double nextBreak(device_ *r, state_ *state, double time)
{
	devicePrivate_ *p;
	double v1, v2, tr, td, tf, pw, per, td1, td2, tb = -1.0, tp;
	
	ReturnNaNIf(r == NULL);
	ReturnNaNIf(time < 0.0);
	p = r->private;
	ReturnNaNIf(p == NULL);
	
	switch(p->srcType) {
	case SOURCE_NONE:
		return 0;
	case SOURCE_PULSE:
		v1 = p->srcData[PULSE_V1];
		v2 = p->srcData[PULSE_V2];
		td = p->srcData[PULSE_TD];
		tr = p->srcData[PULSE_TR];
		tf = p->srcData[PULSE_TF];
		pw = p->srcData[PULSE_PW];
		per = p->srcData[PULSE_PER];
		tp = fmod(time, per);
		if(tp < td) {
			tb = td - tp + time;
		} else if(tp < (tr + td)) {
			tb = (tr + td) - tp + time;
		} else if(tp < (tr + td + pw)) {
			tb = (tr + td + pw) - tp + time;
		} else if(tp < (tr + td + pw + tf)) {
			tb = (tr + td + pw + tf) - tp + time;
		} else {
			tb = per - tp + time;
		}
		break;
	case SOURCE_SIN:
		td = p->srcData[SIN_TD];
		if(time < td) {
			tb = td;
		}
		/* TODO: add bp at peaks of sin wave? */
		break;
	case SOURCE_EXP:
		td1 = p->srcData[EXP_TD1];
		td2 = p->srcData[EXP_TD2];
		if(time < td1) {
			tb = td1;
		} else if(time < td2) {
			tb = td2;
		}
		break;
	case SOURCE_PWL:
		if((p->pwlCount + PWL_T) < (p->srcLength - 1)) {
			tb = p->srcData[p->pwlCount + PWL_T];
		}
		break;
	case SOURCE_SFFM:
		/* TODO: this wasn't implimented in Spice 3f5, but should be added */
		break;
	}
	
	return tb;
}

inline static int setPulseDefaults(devicePrivate_ *p, state_ *state)
{
	if(p->srcLength < PULSE_LEN) {
		p->srcData[PULSE_PER] = stateGetTStop(state);
		ReturnErrIf(isnan(p->srcData[PULSE_PER]));
		p->srcLength++;
		if(p->srcLength < PULSE_LEN) {
			p->srcData[PULSE_PW] = stateGetTStop(state);
			ReturnErrIf(isnan(p->srcData[PULSE_PW]));
			p->srcLength++;
			if(p->srcLength < PULSE_LEN) {
				p->srcData[PULSE_TF] = stateGetTStep(state);
				ReturnErrIf(isnan(p->srcData[PULSE_TF]));
				p->srcLength++;
				if(p->srcLength < PULSE_LEN) {
					p->srcData[PULSE_TR] = stateGetTStep(state);
					ReturnErrIf(isnan(p->srcData[PULSE_TR]));
					p->srcLength++;
					if(p->srcLength < PULSE_LEN) {
						p->srcData[PULSE_TD] = 0.0;
						p->srcLength++;
						ReturnErrIf(p->srcLength < PULSE_LEN);
					}
				}
			}
		}
	}
	return 0;
}

inline static int setSinDefaults(devicePrivate_ *p, state_ *state)
{
	if(p->srcLength < SIN_LEN) {
		p->srcData[SIN_THETA] = 0.0;
		p->srcLength++;
		if(p->srcLength < SIN_LEN) {
			p->srcData[SIN_TD] = 0.0;
			p->srcLength++;
			if(p->srcLength < SIN_LEN) {
				p->srcData[SIN_FREQ] = 1/stateGetTStop(state);
				ReturnErrIf(isnan(p->srcData[SIN_FREQ]));
				p->srcLength++;
				ReturnErrIf(p->srcLength < SIN_LEN);
			}
		}
	}
	return 0;
}

inline static int setExpDefaults(devicePrivate_ *p, state_ *state)
{
	if(p->srcLength < EXP_LEN) {
		p->srcData[EXP_TAU2] = stateGetTStep(state);
		ReturnErrIf(isnan(p->srcData[EXP_TAU2]));
		p->srcLength++;
		if(p->srcLength < EXP_LEN) {
			p->srcData[EXP_TD2] =  stateGetTStep(state) + 
				((p->srcLength == EXP_TD1) ? 0.0 : p->srcData[EXP_TD1]);
			ReturnErrIf(isnan(p->srcData[EXP_TD2]));
			p->srcLength++;
			if(p->srcLength < EXP_LEN) {
				p->srcData[EXP_TAU1] = stateGetTStep(state);
				ReturnErrIf(isnan(p->srcData[EXP_TAU1]));
				p->srcLength++;
				if(p->srcLength < EXP_LEN) {
					p->srcData[EXP_TD1] = 0.0;
					p->srcLength++;
					ReturnErrIf(p->srcLength < EXP_LEN);
				}
			}
		}
	}
	return 0;
}

inline static int setSFFMDefaults(devicePrivate_ *p, state_ *state)
{
	if(p->srcLength < SFFM_LEN) {
		p->srcData[SFFM_FS] = 1/stateGetTStop(state);
		ReturnErrIf(isnan(p->srcData[SFFM_FS]));
		p->srcLength++;
		if(p->srcLength < SFFM_LEN) {
			p->srcData[SFFM_MDI] = 0.0;
			p->srcLength++;
			if(p->srcLength < SFFM_LEN) {
				p->srcData[SFFM_FC] = 1/stateGetTStop(state);
				ReturnErrIf(isnan(p->srcData[SFFM_FC]));
				p->srcLength++;
				ReturnErrIf(p->srcLength < SFFM_LEN);
			}
		}
	}
	return 0;
}

static int calcStepValue(devicePrivate_ *p, state_ *state, double *value, 
		double time)
{
	double v1, v2, tr, td, tf, pw, per, tn;
	double vo, va, freq, theta, td1, td2;
	double tau1, tau2;
	double fc, mdi, fs;
	
	*value = p->dc;
	
	switch(p->srcType) {
	case SOURCE_NONE:
		return 0;
	case SOURCE_PULSE:
		ReturnErrIf(setPulseDefaults(p, state));
		v1 = p->srcData[PULSE_V1];
		v2 = p->srcData[PULSE_V2];
		td = p->srcData[PULSE_TD];
		tr = p->srcData[PULSE_TR];
		tf = p->srcData[PULSE_TF];
		pw = p->srcData[PULSE_PW];
		per = p->srcData[PULSE_PER];
		tn = fmod(time, per);
		if(tn <= td) {
			*value = v1;
		} else if(tn <= (tr + td)) {
			*value = v1 + ((v2 - v1)/tr)*(tn - td);
		} else if(tn <= (tr + td + pw)) {
			*value = v2;
		} else if(tn <= (tr + td + pw + tf)) {
			*value = v2 - ((v2 - v1)/(tf))*(tn - (tr + td + pw));
		} else {
			*value = v1;
		}
		break;
	case SOURCE_SIN:
		ReturnErrIf(setSinDefaults(p, state));
		vo = p->srcData[SIN_VO];
		va = p->srcData[SIN_VA];
		freq = p->srcData[SIN_FREQ];
		td = p->srcData[SIN_TD];
		theta = p->srcData[SIN_THETA];
		if(time <= td) {
			*value = vo;
		} else {
			*value = vo + va*exp(-(time-td)*theta)*sin(2*M_PI*freq*(time-td));
		}
		break;
	case SOURCE_EXP:
		ReturnErrIf(setExpDefaults(p, state));
		v1 = p->srcData[EXP_V1];
		v2 = p->srcData[EXP_V2];
		td1 = p->srcData[EXP_TD1];
		tau1 = p->srcData[EXP_TAU1];
		td2 = p->srcData[EXP_TD2];
		tau2 = p->srcData[EXP_TAU2];
		if(time <= td1) {
			*value = v1;
		} else if(time <= td2) {
			*value = v1 + (v2-v1)*(1-exp(-(time-td1)/tau1));
		} else {
			*value = v1 + (v2-v1)*(1-exp(-(time-td1)/tau1)) + 
					(v1-v2)*(1-exp(-(time-td2)/tau2));
		}
		break;
	case SOURCE_PWL:
		if(p->pwlCount >= 2) {
			for(; (time < p->srcData[p->pwlCount-PWL_LEN]) && 
				(p->pwlCount < p->srcLength); p->pwlCount -= PWL_LEN);
		}
		for(; (time > p->srcData[p->pwlCount]) && 
			(p->pwlCount < p->srcLength); p->pwlCount += PWL_LEN);
		if(p->pwlCount == 0) {
			*value = p->dc;
		} else if (p->pwlCount >= p->srcLength) {
			*value = p->srcData[p->srcLength-1];
		} else {
			v1 = p->srcData[p->pwlCount - (PWL_LEN - PWL_V)];
			v2 = p->srcData[p->pwlCount + PWL_V];
			td = p->srcData[p->pwlCount - (PWL_LEN - PWL_T)];
			tr = p->srcData[p->pwlCount + PWL_T];
			*value = v1 + ((v2 - v1)/(tr-td))*(time - td);
		}
		break;
	case SOURCE_SFFM:
		ReturnErrIf(setSFFMDefaults(p, state));
		vo = p->srcData[SFFM_VO];
		va = p->srcData[SFFM_VA];
		fc = p->srcData[SFFM_FC];
		mdi = p->srcData[SFFM_MDI];
		fs = p->srcData[SFFM_FS];
		*value = vo + va*sin(2*M_PI*fc*time + mdi*sin(2*M_PI*fs*time));
		break;
	}
	
	return 0;
}

static int stepV(device_ *r, state_ *state, double time, deviceNIType niType)
{
	devicePrivate_ *p;
	double value;
	ReturnErrIf(r == NULL);
	p = r->private;
	ReturnErrIf(p == NULL);
	
	ReturnErrIf(calcStepValue(p, state, &value, time));
	
	ReturnErrIf(rowRHSPlus(p->rowR, value - p->dc));
	p->dc = value;
	
	return 0;
}

static int stepI(device_ *r, state_ *state, double time, deviceNIType niType)
{
	devicePrivate_ *p;
	double value;
	ReturnErrIf(r == NULL);
	p = r->private;
	ReturnErrIf(p == NULL);
	
	ReturnErrIf(calcStepValue(p, state, &value, time));
	
	ReturnErrIf(rowRHSMinus(p->rowK, value - p->dc));
	ReturnErrIf(rowRHSPlus(p->rowJ, value - p->dc));
	p->dc = value;
	
	return 0;
}

static int loadV(device_ *r, state_ *state)
{
	devicePrivate_ *p;
	ReturnErrIf(r == NULL);
	p = r->private;
	ReturnErrIf(p == NULL);
	
	ReturnErrIf(p->rowR == NULL);
	
	Debug("%s Load %p", typeV, r);
	
	/* Modified Nodal Analysis Stamp
	 *	                     	+  /\  -
	 *	  |_Vk_Vj_Ir_|_rhs_|	__/Vr\__
	 *	k | -- --  1 | --  |	k \  / j
	 *	j | -- -- -1 | --  |	   \/
	 *	r |  1 -1 -- | Vr  |	------->
	 *	                    	   Ir
	 */
	
	ReturnErrIf(rowRHSPlus(p->rowR, p->dc));
	
	ReturnErrIf(nodeDataPlus(p->nodeRK, 1.0));
	ReturnErrIf(nodeDataMinus(p->nodeRJ, 1.0));
	ReturnErrIf(nodeDataPlus(p->nodeKR, 1.0));
	ReturnErrIf(nodeDataMinus(p->nodeJR, 1.0));
	
	return 0;
}

static int loadI(device_ *r, state_ *state)
{
	devicePrivate_ *p;
	ReturnErrIf(r == NULL);
	p = r->private;
	ReturnErrIf(p == NULL);
	
	ReturnErrIf(p->rowK == NULL);
	ReturnErrIf(p->rowJ == NULL);
	
	Debug("%s Load %p", typeI, r);
	
	/* Modified Nodal Analysis Stamp
	 *	                  	+  __  -
	 *	  |_Vk_Vj_|_rhs_|	__/  \__
	 *	k | -- -- | -Ir |	k \__/ j
	 *	j | -- -- |  Ir |	------->
	 *	                  	   Ir
	 */
	
	ReturnErrIf(rowRHSMinus(p->rowK, p->dc));
	ReturnErrIf(rowRHSPlus(p->rowJ, p->dc));
	
	return 0;
}

static int unconfig(device_ *r)
{
	devicePrivate_ *p;
	
	ReturnErrIf(r == NULL);
	p = r->private;
	ReturnErrIf(p == NULL);
	
	if(p->srcData != NULL) {
		free(p->srcData);
	}
	
	return 0;
}

int deviceConfigSource(device_ *r, double dc, srcType_ srcType, 
		double *srcData, int srcLength, char type)
{
	devicePrivate_ *p;
	int i;
	
	ReturnErrIf(r == NULL);
	ReturnErrIf(r->private != NULL);
	ReturnErrIf(r->numPins != NUM_PINS);
	
	ReturnErrIf((r->pin[K] == &gndRow) && (r->pin[J] == &gndRow));
	
	switch(srcType) {
	case SOURCE_NONE:
		ReturnErrIf(srcLength != 0);
		break;
	case SOURCE_PULSE:
		ReturnErrIf(srcLength != PULSE_LEN);
		break;
	case SOURCE_SIN:
		ReturnErrIf(srcLength != SIN_LEN);
		break;
	case SOURCE_EXP:
		ReturnErrIf(srcLength != EXP_LEN);
		break;
	case SOURCE_PWL:
		ReturnErrIf(MOD(srcLength, PWL_LEN) != 0, 
			"Data points must be a multiple of 2 but you have %i", srcLength);
		for(i = PWL_LEN; i < srcLength; i += PWL_LEN) {
			ReturnErrIf(srcData[i+PWL_T] < srcData[i-(PWL_LEN-PWL_T)], 
					"PWL must have increasing time points %e < %e",
					srcData[i+PWL_T], srcData[i-(PWL_LEN-PWL_T)]);
		}
		break;
	case SOURCE_SFFM:
		ReturnErrIf(srcLength != SFFM_LEN);
		break;
	}
	
	switch(type) {
	case 'v':
		Debug("Configuring %s %p", typeV, r);
		r->type = typeV;
		r->load = loadV;
		r->step = stepV;
		break;
	case 'i':
		Debug("Configuring %s %p", typeI, r);
		r->type = typeI;
		r->load = loadI;
		r->step = stepI;
		break;
	default:
		ReturnErr("Type should be i ro v");
	}
	
	r->unconfig = unconfig;
	r->nextBreak = nextBreak;
	r->private =  calloc(1, sizeof(devicePrivate_));
	ReturnErrIf(r->private == NULL);
	p = r->private;
	
	p->dc = dc;
	p->srcType = srcType;
	p->srcData = srcData;
	p->srcLength = srcLength;
	p->pwlCount = 0;
	
	switch(type) {
	case 'v':
		p->rowR = matrixFindRow(r->matrix, r->refdes, TYPE_EXTENDED);
		ReturnErrIf(p->rowR == NULL);
		p->nodeRK = matrixFindNode(r->matrix, p->rowR, r->pin[K]);
		ReturnErrIf(p->nodeRK == NULL);
		p->nodeRJ = matrixFindNode(r->matrix, p->rowR, r->pin[J]);
		ReturnErrIf(p->nodeRJ == NULL);
		p->nodeKR = matrixFindNode(r->matrix, r->pin[K], p->rowR);
		ReturnErrIf(p->nodeKR == NULL);
		p->nodeJR = matrixFindNode(r->matrix, r->pin[J], p->rowR);
		ReturnErrIf(p->nodeJR == NULL);
		break;
	case 'i':
		p->rowK = r->pin[K];
		ReturnErrIf(p->rowK == NULL);
		p->rowJ = r->pin[J];
		ReturnErrIf(p->rowJ == NULL);
		break;
	default:
		ReturnErr("Type should be i ro v");
	}
	
	return 0;
}
